# -*- coding: utf-8 -*-
#
# Copyright (c) 2019~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

from .ioc import ScopedServiceProvider, LifeTime
from .symbols import Symbols, _Symbol
from .utils import inject_by_anno, inject_by_name

inject_by_table = {
    'anno': inject_by_anno,
    'inject_by_anno': inject_by_anno,
    'name': inject_by_name,
    'inject_by_name': inject_by_name
}

class ServiceProviderBuilder:
    '''
    a decorator api helper class for `ServiceProvider`.
    '''
    __slots__ = ('_provider')

    def __init__(self, provider: ScopedServiceProvider):
        self._provider = provider

    def _on_key_added(self, key):
        pass

    def register(self, lifetime: LifeTime, key=None, factory=None, *, inject_by=None):
        '''
        Register a service factory by `key` with `lifetime`,
        can be use as decorator if the `factory` set `None`.

        If the `key` is `None`, use the `factory` as key. This is helpful to use a `type` as factory.
        Special, if the `key` is a `list`, use each item in it as key;

        By default, the parameter `factory` accept a callable like `() -> Any` or `(ioc) -> Any`,
        but you can pass a `inject_by` function to wrap the `factory` to `(ioc) -> Any`.
        `anyioc` include some useful `inject_by_*` helper functions in `anyioc.utils`.

        returns:

        - decorator mode return the factory.
        - non-decorator mode always return None
        '''

        if isinstance(inject_by, str):
            if inject_by in inject_by_table:
                inject_by = inject_by_table[inject_by]
            else:
                raise ValueError(f'{inject_by} is not one of known `inject by` action.')

        def decorator(func):
            aliases = ()
            if key is None:
                safe_key = func
            elif isinstance(key, list):
                if not key:
                    raise ValueError('cannot use empty list as key list')
                safe_key = _Symbol(f'ref-for-{key[0]}')
                aliases = key
            else:
                safe_key = key

            wraped_func = inject_by(func) if inject_by else func
            self._provider.register(safe_key, wraped_func, lifetime)
            self._on_key_added(safe_key)
            for k in aliases:
                self._provider.register_bind(k, safe_key)
            return func

        if factory is None:
            return decorator
        else:
            decorator(factory)

    def singleton(self, key=None, factory=None, *, inject_by=None):
        '''
        Call `register()` with `LifeTime.singleton`.

        See `register()` for more details.
        '''
        return self.register(LifeTime.singleton, key, factory, inject_by=inject_by)

    def scoped(self, key=None, factory=None, *, inject_by=None):
        '''
        Call `register()` with `LifeTime.scoped`.

        See `register()` for more details.
        '''
        return self.register(LifeTime.scoped, key, factory, inject_by=inject_by)

    def transient(self, key=None, factory=None, *, inject_by=None):
        '''
        Call `register()` with `LifeTime.transient`.

        See `register()` for more details.
        '''
        return self.register(LifeTime.transient, key, factory, inject_by=inject_by)

    def value(self, key):
        '''
        get a decorator that use to register a singleton value by key.
        '''

        def decorator(value):
            self._provider.register_value(key, value)
            self._on_key_added(key)
            return value

        return decorator

    def group(self, *keys):
        '''
        add a new group into `ServiceProvider` by keys,
        return a group scoped `ServiceProviderBuilder` to build sub services.

        the return instance can also use like a unique key.
        '''
        from .utils import make_group

        group = Group(self._provider)
        self._provider.register_group(group, group)
        self._on_key_added(group)
        for k in keys:
            self._provider.register_bind(k, group)
        return group


class Group(ServiceProviderBuilder):
    '''
    a `Group` class use for build `ServiceProvider` and use as unique group key.
    '''
    __slots__ = ('_group_keys_list', )

    def __init__(self, provider: ScopedServiceProvider):
        super().__init__(provider)
        self._group_keys_list = []

    def __iter__(self):
        return iter(self._group_keys_list)

    def _on_key_added(self, key):
        self._group_keys_list.append(key)
        super()._on_key_added(key)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

from typing import Dict, List, Any, Iterable

class ServicesMap:
    def __init__(self, *maps):
        self.maps: List[Dict[Any, list]] = list(maps) or [{}]

    def resolve(self, key):
        '''
        resolve items with reverse order.
        '''
        for mapping in self.maps:
            yield from reversed(mapping.get(key, []))

    def __setitem__(self, key, value):
        self.maps[0].setdefault(key, []).append(value)

    def __getitem__(self, key):
        'get item or raise `KeyError`` if not found'
        for value in self.resolve(key):
            return value
        raise KeyError(key)

    def get(self, key, default=None):
        'get item or `default` if not found'
        for value in self.resolve(key):
            return value
        return default

    def get_many(self, key):
        'get items as list'
        return list(self.resolve(key))

    def scope(self):
        return self.__class__({}, *self.maps)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

from abc import abstractmethod
from typing import Any, List
from contextlib import ExitStack, nullcontext
from threading import RLock
from types import MappingProxyType
from logging import getLogger

from .err import ServiceNotFoundError
from .symbols import Symbols
from ._servicesmap import ServicesMap
from .ioc_resolver import IServiceInfoResolver, ServiceInfoChainResolver
from .ioc_service_info import (
    LifeTime,
    IServiceInfo,
    ServiceInfo,
    ProviderServiceInfo,
    GetAttrServiceInfo,
    ValueServiceInfo,
    GroupedServiceInfo,
    BindedServiceInfo,
    CallerFrameServiceInfo
)

_logger = getLogger(__name__)


class IServiceProvider:
    '''
    the base interface for `ServiceProvider`.
    '''

    @abstractmethod
    def __getitem__(self, key):
        raise NotImplementedError

    @abstractmethod
    def get(self, key, d=None) -> Any:
        '''
        get a service by key.
        '''
        raise NotImplementedError

    @abstractmethod
    def get_many(self, key) -> List[Any]:
        '''
        get services by key.
        '''
        raise NotImplementedError

    @abstractmethod
    def scope(self):
        '''
        create a scoped service provider for get scoped services.
        '''
        raise NotImplementedError


class ScopedServiceProvider(IServiceProvider):

    def __init__(self, services: ServicesMap, parent=None):
        super().__init__()
        self._services = services
        self._exit_stack = None
        self._scoped_cache = {}
        self._parent = parent
        self.__class__ = ServiceProvider # hack
        if parent is None: # root provider
            self._lock = RLock()
        else:
            self._lock = nullcontext()

    def _get_service_info(self, key) -> IServiceInfo:
        try:
            return self._services[key]
        except KeyError:
            pass
        # load missing resolver and resolve service info.
        resolver: IServiceInfoResolver = self._services[Symbols.missing_resolver].get(self)
        return resolver.get(self, key)

    def _get_service_info_list(self, key) -> List[IServiceInfo]:
        return self._services.get_many(key)

    def __getitem__(self, key):
        _logger.debug('get service by key: %r', key)
        service_info = self._get_service_info(key)
        try:
            return service_info.get(self)
        except ServiceNotFoundError as err:
            raise ServiceNotFoundError(key, *err.resolve_chain)

    def get(self, key, d=None) -> Any:
        '''
        get a service by key.
        '''
        try:
            return self[key]
        except ServiceNotFoundError as err:
            if len(err.resolve_chain) == 1:
                return d
            raise

    def get_many(self, key) -> List[Any]:
        '''
        get services by key.

        ### example

        when you registered multi services with the same key,
        you can get them all:

        ``` py
        provider.register_value('a', 1)
        provider.register_value('a', 2)
        assert provider.get_many('a') == [2, 1] # rev order
        ```
        '''
        _logger.debug('get services by key: %r', key)
        service_infos = self._get_service_info_list(key)
        try:
            return [si.get(self) for si in service_infos]
        except ServiceNotFoundError as err:
            raise ServiceNotFoundError(key, *err.resolve_chain)

    def enter(self, context):
        '''
        enter the context.

        returns the result of the `context.__enter__()` method.
        '''
        with self._lock:
            if self._exit_stack is None:
                self._exit_stack = ExitStack()
            return self._exit_stack.enter_context(context)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        with self._lock:
            if self._exit_stack is not None:
                self._exit_stack.__exit__(*args)
                self._exit_stack = None

    def register_service_info(self, key, service_info: IServiceInfo):
        '''
        register a `IServiceInfo` by key.
        '''
        if not isinstance(service_info, IServiceInfo):
            raise TypeError('service_info must be instance of IServiceInfo.')
        _logger.debug('register %r with key %r', service_info, key)
        self._services[key] = service_info

    def register(self, key, factory, lifetime):
        '''
        register a service factory by key.

        `factory` accept a function which require one or zero parameter.
        if the count of parameter is 1, pass a `IServiceProvider` as the argument.
        '''
        return self.register_service_info(key, ServiceInfo(self, key, factory, lifetime))

    def register_singleton(self, key, factory):
        '''
        register a service factory by key.

        `factory` accept a function which require one or zero parameter.
        if the count of parameter is 1, pass a `IServiceProvider` as the argument.
        '''
        return self.register(key, factory, LifeTime.singleton)

    def register_scoped(self, key, factory):
        '''
        register a service factory by key.

        `factory` accept a function which require one or zero parameter.
        if the count of parameter is 1, pass a `IServiceProvider` as the argument.
        '''
        return self.register(key, factory, LifeTime.scoped)

    def register_transient(self, key, factory):
        '''
        register a service factory by key.

        `factory` accept a function which require one or zero parameter.
        if the count of parameter is 1, pass a `IServiceProvider` as the argument.
        '''
        return self.register(key, factory, LifeTime.transient)

    def register_value(self, key, value):
        '''
        register a value by key.

        equals `register_transient(key, lambda ioc: value)`
        '''
        return self.register_service_info(key, ValueServiceInfo(value))

    def register_group(self, key, keys: list):
        '''
        register a grouped `key` for get other `keys`.

        the `keys` can be a ref and you can update it later.

        for example:

        ``` py
        provider.register_value('str', 'name')
        provider.register_value('int', 1)
        provider.register_group('any', ['str', 'int'])
        assert provider['any'] == ('name', 1)
        ```

        equals `register_transient(key, lambda ioc: tuple(ioc[k] for k in keys))`
        '''
        return self.register_service_info(key, GroupedServiceInfo(keys))

    def register_bind(self, new_key, target_key):
        '''
        bind `new_key` to `target_key` so
        you can use `new_key` as key to get value from service provider.

        equals `register_transient(new_key, lambda ioc: ioc[target_key])`
        '''
        return self.register_service_info(new_key, BindedServiceInfo(target_key))

    def scope(self):
        '''
        create a scoped service provider.
        '''
        sp = ScopedServiceProvider(self._services.scope(), self)
        self.enter(sp)
        return sp

    @property
    def builder(self):
        '''
        get a new `ServiceProviderBuilder` wrapper for this `ServiceProvider`.
        '''
        from .builder import ServiceProviderBuilder
        return ServiceProviderBuilder(self)


class ServiceProvider(ScopedServiceProvider):
    '''
    the default impl for `IServiceProvider`.
    '''

    def __init__(self, auto_enter=False):
        super().__init__(ServicesMap())
        provider_service_info = ProviderServiceInfo()
        self._services[Symbols.provider] = provider_service_info
        self._services[Symbols.provider_root] = ValueServiceInfo(self)
        self._services[Symbols.provider_parent] = GetAttrServiceInfo('_parent')
        self._services[Symbols.cache] = GetAttrServiceInfo('_scoped_cache')
        self._services[Symbols.missing_resolver] = ValueServiceInfo(ServiceInfoChainResolver())
        self._services[Symbols.caller_frame] = CallerFrameServiceInfo()

        # service alias
        self._services['ioc'] = provider_service_info
        self._services['provider'] = provider_service_info
        self._services['service_provider'] = provider_service_info
        self._services[ServiceProvider] = provider_service_info
        self._services[IServiceProvider] = provider_service_info

        # options
        self._services[Symbols.provider_options] = ValueServiceInfo(MappingProxyType(
            dict(
                auto_enter=auto_enter
            )
        ))



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020~2999 - Cologler <skyoflw@gmail.com>
# ----------
# the internal utils.
# user should not import anything from this file.
# ----------

import sys
import atexit
import inspect

def get_module_name(fr: inspect.FrameInfo):
    'get module name from frame info'
    mo = inspect.getmodule(fr.frame)
    name = '<stdin>' if mo is None else mo.__name__
    return name

def dispose_at_exit(provider):
    '''
    register `provider.__exit__()` into `atexit` module.

    return the `provider` itself.
    '''
    @atexit.register
    def provider_dispose_at_exit():
        provider.__exit__(*sys.exc_info())
    return provider



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

import sys
from typing import List, Tuple, Union, Any, Dict, Callable
import inspect
from inspect import signature, Parameter

from ._utils import get_module_name

def update_wrapper(wrapper, wrapped):
    ' update wrapper with internal attributes. '
    wrapper.__anyioc_wrapped__ = getattr(wrapped, '__anyioc_wrapped__', wrapped)
    return wrapper


def injectable(*pos_args: List[Union[Tuple[Any], Tuple[Any, Any]]],
               **kw_args: Dict[str, Union[Tuple[Any], Tuple[Any, Any]]]):
    '''
    return a decorator that use to wrap a callable with signature `(ioc) => any`.

    for each item in `pos_args` and each value in `kw_args`:

    - must be a tuple, with 1 or 2 elements;
    - element 1 is the key for get service from `ServiceProvider` container;
    - element 2 is the default value if provide,
      otherwise will use `ServiceProvider.__getitem__()` to get service;
    -

    ### Example:

    ``` py
    @injectable(a=('key1', 1), b=('key2'))
    def func(a, b):
        return a + b
    ```

    is equals:

    ``` py
    def func(ioc):
        def _func(a, b):
            return a + b
        return _func(a=ioc.get('key1', 1), b=ioc['key2'])
    ```
    '''
    for tup in list(pos_args) + list(kw_args.values()):
        if not isinstance(tup, tuple):
            raise TypeError(f'excepted tuple, got {type(tup)}')
        if len(tup) not in (1, 2):
            raise ValueError('tuple should contains 1 or 2 elements')

    def decorator(func):
        def new_func(ioc):
            args = []
            for item in pos_args:
                if len(item) == 1:
                    args.append(ioc[item[0]])
                else:
                    key, default = item
                    args.append(ioc.get(key, default))
            kwargs = {}
            for name, item in kw_args.items():
                if len(item) == 1:
                    kwargs[name] = ioc[item[0]]
                else:
                    key, default = item
                    kwargs[name] = ioc.get(key, default)
            return func(*args, **kwargs)
        return update_wrapper(new_func, func)

    return decorator

def inject_by_key_selector(selector: Callable[[Parameter], Any]):
    '''
    inject arguments by `selector`.

    return a decorator that use to wrap a callable with signature `(ioc) => any`.

    `selector` should be a callable which accept a `inspect.Parameter` object as argument,
    return the key use for inject.

    Note: var keyword parameters and var positional parameters will be ignore.
    '''

    if not callable(selector):
        raise TypeError

    def decorator(func):
        sign = signature(func)
        params = [p for p in sign.parameters.values()]
        pos_args = []
        kw_args = {}
        for param in params:
            ioc_key = selector(param)
            val = (ioc_key, ) if param.default is Parameter.empty else (ioc_key, param.default)
            if param.kind in (Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD):
                pos_args.append(val)
            elif param.kind == Parameter.KEYWORD_ONLY:
                kw_args[param.name] = val
        return injectable(*pos_args, **kw_args)(func)

    return decorator

def inject_by_name(func=None):
    '''
    wrap a callable with signature `(ioc) => any` for inject arguments by parameter name.

    return a decorator when `func` is `None`.

    ### Example:

    ``` py
    @inject_by_name
    def func(a, b='x'):
        return a + b
    ```

    is equals:

    ``` py
    def func(ioc):
        def _func(a, b):
            return a + b
        return _func(a=ioc['a'], b=ioc.get('b', 'x'))
    ```
    '''
    decorator = inject_by_key_selector(lambda x: x.name)

    return decorator if func is None else decorator(func)

def inject_by_anno(func=None, *, use_name_if_empty: bool = False):
    '''
    wrap a callable with signature `(ioc) => any` for inject arguments by parameter annotation.

    return a decorator when `func` is `None`.

    Options:

    - `use_name_if_empty`: whether use `Parameter.name` as key when the `Parameter.annotation` is empty.

    ### Example:

    ``` py
    @inject_by_anno
    def func(a: int, b: str='x'):
        return a + b
    ```

    is equals:

    ``` py
    def func(ioc):
        def _func(a, b):
            return a + b
        return _func(a=ioc[int], b=ioc.get(str, 'x'))
    ```
    '''
    def decorator(func):
        def selector(param: Parameter):
            anno = param.annotation
            if anno is Parameter.empty:
                if use_name_if_empty:
                    ioc_key = param.name
                elif param.default is Parameter.empty:
                    raise ValueError(f'annotation of args {param.name} is empty.')
                else:
                    # use `object()` as key to ensure never get any value from container.
                    ioc_key = object()
            else:
                ioc_key = anno
            return ioc_key

        return inject_by_key_selector(selector)(func)

    return decorator if func is None else decorator(func)

def inject_by_keys(**keys):
    '''
    inject arguments by `keys`.

    return a decorator that use to wrap a callable with signature `(ioc) => any`.

    - keys should be parameter name of func.
    - values are the key that use to get service from service provider.

    ### Example:

    ``` py
    @inject_by_keys(a='key1', b='key2')
    def func(a, b):
        return a + b
    ```

    is equals:

    ``` py
    def func(ioc):
        def _func(a, b):
            return a + b
        return _func(a=ioc['key1'], b=ioc['key2'])
    ```
    '''

    kw_args = dict((k, (v, )) for k, v in keys.items())
    return injectable(**kw_args)

def auto_enter(func):
    '''
    auto enter the context manager when it created.

    the signature of func should be `(ioc) => any`.
    '''
    def new_func(ioc):
        item = func(ioc)
        ioc.enter(item)
        return item
    return new_func

def make_group(container, group_key=None):
    '''
    add a new group into `container` by key `group_key`.
    if `group_key` is `None`, use return function as key.

    return a function accept single argument for add next group item key.
    '''
    group_keys = []

    def add_next_key(next_group_key):
        '''
        add next key into this group.
        '''
        group_keys.append(next_group_key)
        return next_group_key

    if group_key is None:
        group_key = add_next_key

    container.register_group(group_key, group_keys)

    return add_next_key

def get_logger(ioc):
    '''
    a helper that use to get logger from ioc.

    Usage:

    ``` py
    ioc.register_transient('logger', get_logger) # use transient to ensure no cache
    logger = ioc['logger']
    assert logger.name == __name__ # the logger should have module name
    ```
    '''
    import logging
    from .symbols import Symbols

    fr = ioc[Symbols.caller_frame]
    name = get_module_name(fr)
    return logging.getLogger(name)


class Releaser:
    '''
    A helper class to wrap a function as a exitable object.
    The `callback` will be call when the context manager exit.
    '''
    __slots__ = '_callback'

    def __init__(self, callback) -> None:
        self._callback = callback

    def __enter__(self):
        return None

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._callback()

# keep old func names:

auto_inject = inject_by_name



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------


class _Symbol:
    'a symbol _Symbol with description'

    __slots__ = ('_name', )

    def __init__(self, name=''):
        self._name = name

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f'Symbol({self._name})'


class Symbols:
    '''
    the symbols use for ioc.

    this keys are predefined in `ServiceProvider`.
    overwrite this keys will break the expected behavior.
    '''

    # current scoped `IServiceProvider`
    provider = _Symbol('provider')

    # the root `IServiceProvider`
    provider_root = _Symbol('provider_root')

    # the parent of current `IServiceProvider`
    provider_parent = _Symbol('provider_parent')

    # the cache dict from `IServiceProvider`
    cache = _Symbol('cache')

    # the missing resolver from `IServiceProvider`
    missing_resolver = _Symbol('missing_resolver')

    # get frame info of caller
    caller_frame = _Symbol('caller_frame')

    # the options for the `ServiceProvider`, value is a dict like object.
    provider_options = _Symbol('provider_options')



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

from abc import abstractmethod, ABC
from enum import Enum
from inspect import signature, Parameter
from typing import Any
from threading import RLock
import inspect

from .symbols import Symbols
from .utils import update_wrapper


class LifeTime(Enum):
    transient = 0
    scoped = 1
    singleton = 2


class IServiceInfo(ABC):
    __slots__ = ()

    @abstractmethod
    def get(self, provider) -> Any:
        raise NotImplementedError


class ServiceInfo(IServiceInfo):
    '''generic `IServiceInfo`.'''

    __slots__ = (
        '_key', '_lifetime', '_factory', '_factory_origin',
        # for not transient
        '_lock',
        # for singleton
        '_cache_value', '_service_provider',
        # options
        '_options',
    )

    def __init__(self, service_provider, key, factory, lifetime):
        self._factory_origin = factory

        sign = signature(factory)
        if not sign.parameters:
            self._factory = update_wrapper(lambda _: factory(), factory)
        elif len(sign.parameters) == 1:
            arg_0 = list(sign.parameters.values())[0]
            if arg_0.kind in (Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD):
                self._factory = factory
            elif arg_0.kind == Parameter.KEYWORD_ONLY:
                arg_0_name = arg_0.name
                self._factory = update_wrapper(lambda _arg: factory(**{arg_0_name: _arg}), factory)
            else:
                raise ValueError(f'unsupported factory signature: {sign}')
        else:
            raise TypeError('factory has too many parameters.')

        self._key = key
        self._lifetime = lifetime
        self._cache_value = None
        self._service_provider = service_provider
        self._options: dict = service_provider[Symbols.provider_options]

        if self._lifetime != LifeTime.transient:
            self._lock = RLock()
        else:
            self._lock = None

        if self._lifetime == LifeTime.singleton:
            # service_provider is required when lifetime == singleton
            assert self._service_provider is not None

    def __repr__(self) -> str:
        return f'<Service: {self._lifetime}, {self._factory_origin!r}>'

    def get(self, provider):
        if self._lifetime is LifeTime.transient:
            return self._create(provider)

        if self._lifetime is LifeTime.scoped:
            return self._from_scoped(provider)

        if self._lifetime is LifeTime.singleton:
            return self._from_singleton()

        raise NotImplementedError(f'what is {self._lifetime}?')

    def _from_scoped(self, provider):
        cache: dict = provider[Symbols.cache]
        try:
            return cache[self]
        except KeyError:
            pass
        with self._lock:
            try:
                return cache[self]
            except KeyError:
                service = self._create(provider)
                cache[self] = service
                return service

    def _from_singleton(self):
        if self._cache_value is None:
            with self._lock:
                if self._cache_value is None:
                    self._cache_value = (self._create(self._service_provider), )
        return self._cache_value[0]

    def _create(self, provider):
        '''
        return the finally service instance.
        '''

        service = self._factory(provider)
        if self._options['auto_enter']:
            wrapped = getattr(self._factory, '__anyioc_wrapped__', self._factory)
            if isinstance(wrapped, type) and hasattr(wrapped, '__enter__') and hasattr(wrapped, '__exit__'):
                service = provider.enter(service)
        return service


class ProviderServiceInfo(IServiceInfo):
    '''a `IServiceInfo` use for get current `ServiceProvider`.'''

    __slots__ = ()

    def __repr__(self) -> str:
        return f'<Provider>'

    def get(self, provider):
        return provider


class GetAttrServiceInfo(IServiceInfo):
    '''getattr from current `ServiceProvider`.'''

    __slots__ = ('_attr_info')

    def __init__(self, *attr_info: tuple):
        super().__init__()
        self._attr_info = attr_info

    def __repr__(self) -> str:
        return f'<GetAttr: {self._attr_info[0]!r}>'

    def get(self, provider):
        return getattr(provider, *self._attr_info)


class ValueServiceInfo(IServiceInfo):
    '''a `IServiceInfo` use for get fixed value.'''

    __slots__ = ('_value')

    def __init__(self, value):
        self._value = value

    def __repr__(self) -> str:
        return f'<Value: {self._value!r}>'

    def get(self, provider):
        return self._value


class GroupedServiceInfo(IServiceInfo):
    '''a `IServiceInfo` use for get multi values as a tuple from keys list.'''

    __slots__ = ('_keys')

    def __init__(self, keys: list):
        self._keys = keys

    def get(self, provider):
        return tuple(provider[k] for k in self._keys)


class BindedServiceInfo(IServiceInfo):
    '''a `IServiceInfo` use for get value from target key.'''

    __slots__ = ('_target_key')

    def __init__(self, target_key):
        self._target_key = target_key

    def __repr__(self) -> str:
        return f'<Binded: {self._target_key!r}>'

    def get(self, provider):
        return provider[self._target_key]


class CallerFrameServiceInfo(IServiceInfo):
    'a `IServiceInfo` use for get caller frameinfo'

    __slots__ = ()

    def get(self, _):
        frs = inspect.getouterframes(inspect.currentframe())
        for fr in frs[2:]:
            mo = inspect.getmodule(fr.frame)
            if mo is None or mo.__name__.partition('.')[0] != 'anyioc':
                return fr



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['anyioc']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'anyioc',
    'version': '0.12.1',
    'description': 'Another simple ioc framework for python',
    'long_description': "# anyioc\n\n![GitHub](https://img.shields.io/github/license/Cologler/anyioc-python.svg)\n[![CI](https://github.com/Cologler/anyioc-python/actions/workflows/main.yml/badge.svg)](https://github.com/Cologler/anyioc-python/actions/workflows/main.yml)\n[![PyPI](https://img.shields.io/pypi/v/anyioc.svg)](https://pypi.org/project/anyioc/)\n![GitHub code size in bytes](https://img.shields.io/github/languages/code-size/Cologler/anyioc-python)\n\nAnother simple ioc framework for python.\n\n## Usage\n\n``` py\nfrom anyioc import ServiceProvider\nprovider = ServiceProvider()\nprovider.register_singleton('the key', lambda ioc: 102) # ioc will be a `IServiceProvider`\nvalue = provider.get('the key')\nassert value == 102\n```\n\n## Register and resolve\n\nBy default, you can use following methods to register services:\n\n- `ServiceProvider.register_singleton(key, factory)`\n- `ServiceProvider.register_scoped(key, factory)`\n- `ServiceProvider.register_transient(key, factory)`\n- `ServiceProvider.register(key, factory, lifetime)`\n- `ServiceProvider.register_value(key, value)`\n- `ServiceProvider.register_group(key, keys)`\n- `ServiceProvider.register_bind(new_key, target_key)`\n\nAnd use following methods to resolve services:\n\n- `ServiceProvider.__getitem__(key)`\n- `ServiceProvider.get(key)`\n- `ServiceProvider.get_many(key)`\n\n*`get` return `None` if the service was not found, but `__getitem__` will raise a `ServiceNotFoundError`.*\n\nRead full [documentation](https://github.com/Cologler/anyioc-python/wiki).\n",
    'author': 'Cologler',
    'author_email': 'skyoflw@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

from .ioc import IServiceProvider, ServiceProvider, ServiceNotFoundError
from .ioc_service_info import LifeTime

__all__ = [
    'IServiceProvider',
    'ServiceProvider',
    'ServiceNotFoundError',
    'LifeTime',
]



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

from importlib import import_module

from .ioc import ServiceProvider
from .ioc_service_info import LifeTime
from .builder import inject_by_table
from .utils import inject_by_keys


class BadConfError(Exception):
    '''
    raise when the conf dict was incorrect.
    '''

def _iter_list_with_path(path: str, l: list):
    'yield a (path, item) tuple from list.'
    for i, v in enumerate(l):
        yield f'{path}[{i!r}]', v

def _iter_dict_with_path(path: str, d: dict):
    'yield a (path, key, value) tuple from dict.'
    for k, v in d.items():
        yield f'{path}[{k!r}]', k, v

def _ensure_is_dict(path: str, obj, fields=()):
    if not isinstance(obj, dict):
        raise BadConfError(f'<{path}> must be a dict.')
    if fields:
        for field in fields:
            if field not in obj:
                raise BadConfError(f'<{path}> miss {field!r} field.')

def _getattr_from_module(path: str, mod, fullname: str):
    names = fullname.split('.')
    assert names
    if not all(n for n in names):
        raise BadConfError(f'name of <{path}> contain empty part.')
    try:
        attr = mod
        for name in names:
            attr = getattr(attr, name)
        return attr
    except AttributeError:
        raise BadConfError(f'<{path}>: no such attr {fullname!r} on module {mod.__name__!r}.')

def _load_object(path: str, conf, allow_module=False):
    '''
    load a object from runtime.

    if `aallow_module` is `True`, that mean allow the return object to be a module.
    '''
    module_name = None
    object_name = None

    if isinstance(conf, str):
        conf: str
        # parse like console_scripts
        parts = conf.split(':')
        if allow_module:
            if len(parts) > 2:
                raise BadConfError(
                    f'<{path}> should be a `module-name[:callable-name]` like str.')
            module_name, object_name = (parts + [None, None])[:2]
        else:
            if len(parts) != 2:
                raise BadConfError(
                    f'<{path}> should be a `module-name:callable-name` like str.')
            module_name, object_name = parts
            if not object_name:
                raise BadConfError(f'name part of <{path}/name> is empty.')
        if not module_name:
            raise BadConfError(f'module part of <{path}/module> is empty.')

    elif isinstance(conf, dict):
        conf: dict
        module_name = conf.get('module', module_name)
        object_name = conf.get('name', object_name)
        if not isinstance(module_name, str):
            raise BadConfError(f'<{path}/module> is not a str.')
        if not module_name:
            raise BadConfError(f'<{path}/module> is empty.')
        if not allow_module:
            if not isinstance(object_name, str):
                raise BadConfError(f'<{path}/name> is not a str.')
            if not object_name:
                raise BadConfError(f'<{path}/name> is empty.')

    else:
        raise BadConfError(f'<{path}> is not either str or dict.')

    try:
        rv = import_module(module_name)
    except ImportError:
        raise BadConfError(f'<{path}>: unable import module {module_name!r}.')

    if object_name:
        name_parts = object_name.split('.')
        if not all(n for n in name_parts):
            raise BadConfError(f'<{path}> contain empty part.')
        try:
            for name in name_parts:
                rv = getattr(rv, name)
        except AttributeError:
            raise BadConfError(f'<{path}>: no such attr {object_name!r} on module {module_name!r}.')
    return rv

class _ConfLoader:
    def __init__(self, provider: ServiceProvider):
        super().__init__()
        self.provider = provider

    def load_conf(self, conf: dict):
        if not isinstance(conf, dict):
            raise TypeError('conf is not a dict.')

        self._on_services('/services', conf.get('services'))
        self._on_values('/values', conf.get('values'))
        self._on_groups('/groups', conf.get('groups'))
        self._on_binds('/binds', conf.get('binds'))

    def _on_services(self, path: str, services):
        if not services:
            return

        if isinstance(services, dict):
            for p, k, v in _iter_dict_with_path(path, services):
                _ensure_is_dict(p, v)
                key = v.get('key', k)
                if key != k:
                    raise BadConfError(f'<{p}> already contains another key: {key!r}.')
                self._on_service_item(p, v, k)
        elif isinstance(services, list):
            for p, v in _iter_list_with_path(path, services):
                _ensure_is_dict(p, v, ('key', ))
                self._on_service_item(p, v, v['key'])
        else:
            raise BadConfError(f'<{path}> is not either dict or list.')

    def _on_service_item(self, path: str, service_conf: dict, key):
        # factory
        factory = service_conf.get('factory', None)
        if factory is None:
            raise BadConfError(f'<{path}/factory> is null, which is required.')
        elif callable(factory):
            # conf was create in python process instead read from a conf file
            pass
        else:
            factory = _load_object(f'{path}/factory', factory)
            if not callable(factory):
                raise BadConfError(f'<{path}/factory> is not a callable.')

        # lifetime
        lifetime = service_conf.get('lifetime', LifeTime.transient)
        if isinstance(lifetime, LifeTime):
            pass
        else:
            if not isinstance(lifetime, str) or lifetime not in LifeTime.__members__:
                one_of = ', '.join(LifeTime.__members__)
                raise BadConfError(
                    f'value of <{path}/lifetime> ({lifetime!r}) is not one of ({one_of}).')
            lifetime = LifeTime.__members__[lifetime]

        # inject_by
        inject_by = service_conf.get('inject_by', None)
        if inject_by is not None:
            if callable(inject_by):
                # conf was create in python process instead read from a conf file
                pass
            elif isinstance(inject_by, str):
                try:
                    inject_by_decorator = inject_by_table[inject_by]
                except (TypeError, KeyError):
                    # maybe `inject_by` is unhashable.
                    inject_by_decorator = None
                if inject_by_decorator is None:
                    one_of = ', '.join(inject_by_table)
                    raise BadConfError(
                        f'value of <{path}/inject_by> ({inject_by!r}) is not one of ({one_of}).')
                inject_by = inject_by_decorator
            elif isinstance(inject_by, dict):
                keys = {}
                for name in inject_by:
                    if not isinstance(name, str):
                        raise BadConfError(
                            f'key of <{path}/inject_by> must be str.')
                inject_by = inject_by_keys(**inject_by)
            else:
                raise BadConfError(f'<{path}/inject_by> is not either str or dict.')
            factory = inject_by(factory)

        # enter
        enter = bool(service_conf.get('enter', False))

        # register
        def service_factory(ioc: ServiceProvider):
            obj = factory(ioc)
            if enter:
                ioc.enter(obj)
            return obj

        self.provider.register(key, service_factory, lifetime)

    def _on_values(self, path: str, values):
        if not values:
            return

        if isinstance(values, dict):
            for k, v in values.items():
                self.provider.register_value(k, v)
        elif isinstance(values, list):
            for p, c in _iter_list_with_path(path, values):
                self._on_value_item(p, c)
        else:
            raise BadConfError(f'<{path}> is not either dict or list.')

    def _on_value_item(self, path: str, conf: dict):
        fields = ('key', 'value')
        _ensure_is_dict(path, conf, fields)
        key, value = [conf[f] for f in fields]
        if conf.get('ref', False):
            # mean value is a ref from code
            value = _load_object(f'{path}/value', value, allow_module=True)
        self.provider.register_value(key, value)

    def _on_groups(self, path: str, groups):
        if not groups:
            return

        if isinstance(groups, dict):
            for k, v in groups.items():
                if not isinstance(v, list):
                    raise BadConfError(f'value of <{path}[{k!r}]> is not a list.')
                self.provider.register_group(k, v)
        elif isinstance(groups, list):
            fields = ('key', 'keys')
            for p, conf in _iter_list_with_path(path, groups):
                _ensure_is_dict(p, conf, fields)
                k, gk = [conf[f] for f in fields]
                if not isinstance(gk, list):
                    raise BadConfError(f'<{p}/keys> is not a list.')
                self.provider.register_group(k, gk)
        else:
            raise BadConfError(f'<{path}> is not either dict or list.')


    def _on_binds(self, path: str, binds):
        if not binds:
            return

        if isinstance(binds, dict):
            for k, v in binds.items():
                self.provider.register_bind(k, v)
        elif isinstance(binds, list):
            fields = ('key', 'target')
            for p, conf in _iter_list_with_path(path, binds):
                _ensure_is_dict(p, conf, fields)
                values = [conf[f] for f in fields]
                self.provider.register_bind(*values)
        else:
            raise BadConfError(f'<{path}> is not either dict or list.')


def load_conf(provider: ServiceProvider, conf: dict) -> None:
    '''
    load a conf dict and build the `ServiceProvider`.

    if conf dict was incorrect, raise `BadConfError` with useful message.
    '''
    _ConfLoader(provider).load_conf(conf)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018~2999 - Cologler <skyoflw@gmail.com>
# ----------
# a global ioc
# ----------

import importlib
import importlib.util
import functools
import inspect
import threading

from .ioc import ServiceProvider
from .utils import inject_by_name
from ._utils import get_module_name, dispose_at_exit

ioc = ServiceProvider()
dispose_at_exit(ioc)

# scoped global ioc

_module_scoped_providers = {}
_module_scoped_lock = threading.RLock()

def _is_module_exists(module_name: str) -> bool:
    try:
        return importlib.util.find_spec(module_name) is not None
    except ModuleNotFoundError:
        return False

def _get_module_provider(module_name: str):
    'get or create module provider'
    provider = _module_scoped_providers.get(module_name)
    if provider is None:
        with _module_scoped_lock:
            provider = _module_scoped_providers.get(module_name)
            if provider is None:
                provider = ServiceProvider()
                dispose_at_exit(provider)
                _module_scoped_providers[module_name] = provider

                # auto init ioc
                initioc_module_name = module_name + '.init_ioc'
                if _is_module_exists(initioc_module_name):
                    init_ioc = importlib.import_module(initioc_module_name)
                    conf_ioc = getattr(init_ioc, 'conf_ioc', None)
                    if conf_ioc is not None:
                        conf_ioc(provider)
    return provider

def _get_caller_module_name():
    fr = inspect.getouterframes(inspect.currentframe())[2]
    return get_module_name(fr)

def get_module_provider(module_name: str=None) -> ServiceProvider:
    '''
    get the module scoped singleton `ServiceProvider`.

    if `module_name` is `None`, use caller module name.

    if module `{module_name}.init_ioc` exists and it has a attr `conf_ioc`, will auto config like:

    ``` py
    (importlib.import_module(module_name + '.init_ioc')).conf_ioc(module_provider)
    ```
    '''
    if module_name is None:
        module_name = _get_caller_module_name()

    if not isinstance(module_name, str):
        raise TypeError

    return _get_module_provider(module_name)

def get_pkgroot_provider(pkgroot: str=None) -> ServiceProvider:
    '''
    get the package root scoped singleton `ServiceProvider`.

    if `pkgroot` is `None`, use caller package root.

    for example, `get_pkgroot_provider('A.B.C.D')` is equals `get_module_provider('A')`
    '''
    if pkgroot is None:
        pkgroot = _get_caller_module_name()

    if not isinstance(pkgroot, str):
        raise TypeError

    pkgroot = pkgroot.partition('.')[0]
    return _get_module_provider(pkgroot)

# keep old func names:

get_namespace_provider = get_pkgroot_provider



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

from abc import abstractmethod
import typing
from typing import List
from threading import RLock
from contextlib import nullcontext

from .err import ServiceNotFoundError
from .ioc_service_info import ValueServiceInfo, ServiceInfo, LifeTime, IServiceInfo, GroupedServiceInfo

class IServiceInfoResolver:
    '''
    the base class for dynamic resolve `IServiceInfo`.
    '''

    def get(self, provider, key) -> IServiceInfo:
        '''
        get the `IServiceInfo` from resolver.
        '''
        raise ServiceNotFoundError(key)

    def __add__(self, other):
        new_resolver = ServiceInfoChainResolver()
        new_resolver.chain.append(self)
        new_resolver.append(other)
        return new_resolver

    def cache(self, *, sync=False):
        '''
        return a `IServiceInfoResolver` to cache all values from current `IServiceInfoResolver`.
        that mean all values will not dynamic update after first resolved.
        '''
        return CacheServiceInfoResolver(self, sync=sync)


class ServiceInfoChainResolver(IServiceInfoResolver):
    '''
    a helper resolver for resolve values from each `IServiceInfoResolver`
    '''

    def __init__(self, *resolvers):
        self.chain: List[IServiceInfoResolver] = list(resolvers)

    def get(self, provider, key):
        for resolver in self.chain:
            try:
                return resolver.get(provider, key)
            except ServiceNotFoundError:
                pass
        return super().get(provider, key)

    def append(self, other):
        if isinstance(other, ServiceInfoChainResolver):
            self.chain.extend(other.chain)
        else:
            self.chain.append(other)

    def __add__(self, other):
        new_resolver = ServiceInfoChainResolver()
        new_resolver.chain.extend(self.chain)
        new_resolver.append(other)
        return new_resolver


class CacheServiceInfoResolver(IServiceInfoResolver):
    '''
    a helper resolver for cache values from other `IServiceInfoResolver`

    NOTE:
    if a `IServiceInfo` is affect by `provider`, you should not cache it.
    `CacheServiceInfoResolver` only cache by the `key` and ignore the `provider` arguments.
    '''

    def __init__(self, base_resolver: IServiceInfoResolver, *, sync=False):
        super().__init__()
        self._base_resolver = base_resolver
        self._cache = {}
        self._lock = RLock() if sync else nullcontext()

    def get(self, provider, key):
        try:
            return self._cache[key]
        except KeyError:
            pass
        with self._lock:
            try:
                return self._cache[key]
            except KeyError:
                pass
            service_info = self._base_resolver.get(provider, key)
            self._cache[key] = service_info
            return service_info

    def cache(self):
        return self


class ImportServiceInfoResolver(IServiceInfoResolver):
    '''
    dynamic resolve `IServiceInfo` if the key is a package name.
    '''

    def get(self, provider, key):
        import importlib
        if isinstance(key, str):
            try:
                module = importlib.import_module(key)
                return ValueServiceInfo(module)
            except TypeError:
                pass
            except ModuleNotFoundError:
                pass
        return super().get(provider, key)


class SimpleServiceInfo(IServiceInfo):
    __slots__ = ('_factory')

    def __init__(self, factory):
        self._factory = factory

    def get(self, provider):
        return self._factory(provider)


class TypesServiceInfoResolver(IServiceInfoResolver):
    '''
    dynamic resolve `IServiceInfo` if the key is a type instance.
    '''

    inject_by = None

    def get(self, provider, key):
        if isinstance(key, type):
            factory = self.inject_by(key) if self.inject_by else key
            return SimpleServiceInfo(factory)
        return super().get(provider, key)


class TypeNameServiceInfoResolver(IServiceInfoResolver):
    '''
    dynamic resolve `IServiceInfo` if the key is a type name or qualname.
    '''

    inject_by = None

    def _get_type(self, key):
        if isinstance(key, str):
            for klass in object.__subclasses__():
                if getattr(klass, '__name__', None) == key:
                    return klass
                if getattr(klass, '__qualname__', None) == key:
                    return klass
        # None

    def get(self, provider, key):
        klass = self._get_type(str)
        if klass is not None:
            factory = self.inject_by(klass) if self.inject_by else klass
            return SimpleServiceInfo(factory)
        return super().get(provider, key)


class TypingServiceInfoResolver(IServiceInfoResolver):
    '''
    dynamic resolve `IServiceInfo` if the key is a generic type make from typing.

    to use this, you need to install `type_info` module from pypi.
    '''

    get_type_info = None

    def __init__(self) -> None:
        super().__init__()
        if self.get_type_info is None:
            try:
                from type_info import get_type_info
            except ModuleNotFoundError:
                raise ImportError(
                    'You need to install `type_info` module from pypi'
                )
            type(self).get_type_info = staticmethod(get_type_info)

    def get(self, provider, key):
        try:
            type_info = self.get_type_info(key)
        except TypeError:
            type_info = None

        if type_info and type_info.is_generic:
            factory = None

            if type_info.generic_type is typing.Tuple:
                if all(not a.is_generic and not a.is_typevar for a in type_info.generic_args):
                    keys = [t.target_type for t in type_info.generic_args]
                    def factory(ioc):
                        return type_info.std_type(provider[k] for k in keys)

            if factory is not None:
                return SimpleServiceInfo(factory)

        return super().get(provider, key)
